import numpy as np


def costfun(x, r, v_obs, v_err=1.):
    return (r.interpn(x) - v_obs) / v_err


def default_lnlike(x, r, obs, obs_err, obs_weight=None, **kwargs):
    # in case of outlier
    try:
        mod = r(x)
    except Exception:
        return -np.inf

    # chi2
    res2 = ((mod - obs) / (obs_err + r.me)) ** 2.
    if obs_weight is not None:
        res2 *= obs_weight
    if np.all(np.logical_not(np.isfinite(res2))):
        # in case every element is nan
        return -np.inf
    else:
        return - 0.5 * np.nansum(res2)


def best_match(mod, mod_err, obs, obs_err, obs_weight=None, mask=None):
    """ search for the best match (min chi2) template in regli database

    Parameters
    ----------
    mod:
        templates
    mod_err:
        errors of templates
    obs:
        observations
    obs_err:
        errors of observations
    obs_weight:
        weight of observations, np.nan for bad values
    mask:
        1 for bad, 0 for good.

    Return
    ------
    index of the best match template
    """
    res2 = ((mod - obs) / (obs_err + mod_err)) ** 2.
    if obs_weight is not None:
        res2 *= obs_weight
    ind_invalid = np.all(np.logical_not(np.isfinite(res2)), axis=1)
    lnpost = - 0.5 * np.nansum(res2, axis=1)

    if mask is None:
        lnpost = np.ma.MaskedArray(lnpost, ind_invalid)
        return np.nanargmax(lnpost)
    else:
        lnpost = np.ma.MaskedArray(lnpost, mask | ind_invalid)
        return np.nanargmax(lnpost)



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 19 21:11:15 2019

@author: cham
"""

import numpy as np
from .regli import Regli
from slam.normalization import normalize_spectrum, normalize_spectra_block
from scipy.interpolate import interp1d
from astropy import constants as const


SOL = 299792.458  # km/s


class SpIn(object):

    def __init__(self, ):

        self.ndim = 0
        self.nband = 0
        self.wave = list()
        self.reglis = list()

    def add_band(self, wave, flux, params):

        # the new band should have the same dimension
        ndim = params.shape[1]
        if self.nband > 0:
            assert ndim == self.ndim
        else:
            self.ndim = ndim

        # append the new band
        self.wave.append(wave)
        r = Regli.init_from_flats(params)
        r.set_values(flux)
        self.reglis.append(r)
        self.nband += 1

    def interp(self, p, band="all"):
        p = np.array(p)
        if p.ndim == 1:
            p = p.reshape(1, -1)
        # n_p = p.shape[0]

        wave = []
        flux_interp = []

        # bands
        if band == "all":
            band = [_ for _ in range(self.nband)]
        elif isinstance(band, int):
            band = [band]
        else:
            band = list(band)

        # interpolate each band
        for iband in band:
            flux_interp.append(np.array([self.reglis[iband](p_) for p_ in p]))
            wave.append(self.wave[iband])

        return wave, flux_interp

    @staticmethod
    def generate_mock_p_uniform(n=10, ranges=[(0, 1)], seed=0):
        ranges = np.array(ranges)
        ndim_ = len(ranges)
        pmock = np.random.rand(n, ndim_) * np.diff(ranges, axis=1).reshape(1, -1) \
            + ranges[:, 0].reshape(1, -1)
        return pmock

    def generate_mock_flux(self, pmock, mock_snr=50, rv=0, wave_interp=None,
                           fill_flux=-1, fill_ivar=0., norm=True,
                           **norm_kwargs):
        if wave_interp is None:
            wave_interp = self.wave

        nmock = pmock.shape[0]
        # generate mock stellar spectra
        print("@SpIn: interpolate spectra ...")
        wave, mock_flux = self.interp(pmock, band="all")

        mock_z = np.random.randn(nmock) * rv / (const.c.value/1e3)

        mock_flux_rvin = []
        mock_ivar_rvin = []
        print("@SpIn: add noise to spectra ...")
        for iband in range(self.nband):
            # for each band, generate spectra + RV
            this_mock_flux_rvi = np.array([
                interp1d(wave[iband] * (1 + mock_z[i]), mock_flux[iband][i],
                         kind="linear", bounds_error=False,
                         fill_value=np.nan)(wave_interp[iband])
                for i in range(nmock)])

            # check nan
            this_mock_flux_rvi = np.where(np.isfinite(this_mock_flux_rvi),
                                           this_mock_flux_rvi, fill_flux)
            # this_mock_flux_rvin = np.where(np.isfinite(this_mock_flux_rvin),
            #                                this_mock_flux_rvin, fill_flux)

            # simulate S/N
            if mock_snr > 0:
                # add noise
                this_mock_flux_rvin, this_mock_ivar_rvin = add_gaussian_noise_to_spec(
                    this_mock_flux_rvi, snr=mock_snr, fill_ivar=0.)
                mock_flux_rvin.append(this_mock_flux_rvin)
                mock_ivar_rvin.append(this_mock_ivar_rvin)
            else:
                # return original interpolated spec

                mock_flux_rvin.append(this_mock_flux_rvi)
                mock_ivar_rvin.append(this_mock_flux_rvi*0)

        if not norm:
            return mock_flux_rvin, mock_ivar_rvin
        else:
            comb_norm_kwargs = dict(dwave=20,
                                    p=(1e-06, 1e-07),
                                    q=0.6,
                                    eps=1e-10,
                                    rsv_frac=1.0,
                                    n_jobs=1,
                                    verbose=1)
            comb_norm_kwargs.update(norm_kwargs)

            print("@SpIn: normalize spectra ...")
            mock_flux_rvin_norm = []
            mock_ivar_rvin_norm = []
            for iband in range(self.nband):
                this_mock_flux_rvin_norm, this_mock_flux_rvin_cont = \
                    normalize_spectra_block(wave_interp[iband],
                                            mock_flux_rvin[iband],
                                            wave_interp[iband][[0, -1]],
                                            **comb_norm_kwargs)
                mock_flux_rvin_norm.append(this_mock_flux_rvin_norm)
                mock_ivar_rvin_norm.append(mock_ivar_rvin[iband]*this_mock_flux_rvin_cont**2)

            return mock_flux_rvin_norm, mock_ivar_rvin_norm

    def select_finite(self, mock_flux, mock_ivar):
        pass

    def normalize_spectra_block(self, ):
        pass


def add_gaussian_noise_to_spec(spec, snr=100, fill_ivar=0.):
    """ add gaussian noise to spectra """
    specn = (1+np.random.randn(*spec.shape)/snr)*spec
    errn = spec/snr
    ivarn = errn**-2

    indbad = (errn <= 0) | (specn <= 0)
    ivarn = np.where(indbad, fill_ivar, ivarn)
    specn = np.where(indbad, 0., specn)
    return specn, ivarn






# -*- coding: utf-8 -*-
"""
Created on Wed Jun 19 12:30:45 2019

@author: cham
"""

#%%
%pylab

#%%
x = np.linspace(1, 2, 100)
y = x*1.2+np.random.randn(*x.shape)*1+.1
yerr = y*0+.1
                         
#%%
figure()
plot(x, y,'.')

#%%

def f_model(p, x):
    """ p = [a, b] """
    a, b = p
    return a*x+b

def costfun(p, xobs, yobs, yobserr):
    return -np.sum(((f_model(p, xobs)-yobs)/yobserr)**2)/2


#%%
from emcee import EnsembleSampler
sampler = EnsembleSampler(8, 2, costfun, args=[x, y, yerr])
#%%
""" initial positions """
p0 = np.random.randn(8, 2)

""" burn-in """
pos, lnprob, rstate = sampler.run_mcmc(p0, 1000)
figure();plot(sampler.chain[:,:, 0].T)
""" T=transpost"""
""" run mcmc """
sampler.reset()
sampler.run_mcmc(pos, 2000)
figure();plot(sampler.chain[:,:, 0].T)

#%%

from corner import corner
corner(sampler.flatchain)

#%%
def run_mcmc(obs, obs_err=0, mod_err=0, obs_tag=None, p0=None, 
             n_burnin=(100, 100), n_step=1000, 
             lnlike=None, lnprior=None,
             lnlike_kwargs={}, lnprior_kwargs={},
             pos_eps=.1, full=True, shrink="max"):
    """ run MCMC for (obs, obs_err, obs_weight)

    Parameters
    ----------
    obs:
        observable
    obs_err:
        error of observation
    obs_tag:
        array(size(obs), 1 for good, 0 for bad.
    p0:
        initial points
    n_burnin:
        int/sequence, do 1 or multiple times of burn in process
    n_step:
        running step
    lnlike:
        lnlike(x, *args) is the log likelihood function
    lnprior:
        lnpost(x) is the log prior
    pos_eps:
        random magnitude for starting position
    full:
        if True, return sampler, pos, prob, state
        otherwise, return sampler
    shrink:
        default "max": shrink to the maximum likelihood position

    Return
    ------
    EnsembleSampler instance

    """
    if obs_tag is None:
        # default obs_weight
        obs_tag = np.isfinite(obs)
    else:
        obs_tag = np.isfinite(obs) & obs_tag
    # cope with bad obs
    if np.sum(obs_tag) == 0:
        return None

    if p0 is None:
        # do best match for starting position
        p0 = self.best_match(obs, obs_err)
        print("@Regli.best_match: ", p0)

    if lnlike is None:
        # default gaussian likelihood function
        lnlike = default_lnlike
        print("@Regli: using the default gaussian *lnlike* function...")

    if lnprior is None:
        # no prior is specified
        lnpost = lnlike
        print("@Regli: No prior is adopted ...")
    else:
        # use user-defined prior
        print("@Regli: using user-defined *lnprior* function...")

        def lnpost(*_args, **_kwargs):
            lnpost_value = float(
                lnprior(_args[0], **_kwargs["lnprior_kwargs"])) + \
                           lnlike(*_args, **_kwargs["lnlike_kwargs"])
            if np.isfinite(lnpost_value):
                return lnpost_value
            else:
                return -np.inf

    # set parameters for sampler
    ndim = self.ndim
    nwalkers = 2 * ndim

    # initiate sampler
    sampler = EnsembleSampler(nwalkers, ndim, lnpostfn=lnpost,
                              args=(self, obs, obs_err, obs_tag),
                              kwargs=dict(lnprior_kwargs=lnprior_kwargs,
                                          lnlike_kwargs=lnlike_kwargs))

    # generate random starting positions
    pos0 = rand_pos(p0, nwalkers=nwalkers, eps=pos_eps)

    if isinstance(n_burnin, collections.Iterable):
        # [o] multiple burn-ins
        for n_burnin_ in n_burnin:
            # run mcmc
            print("@Regli: running burn-in [{}]...".format(n_burnin_))
            pos, prob, state = sampler.run_mcmc(pos0, n_burnin_)

            # shrink to a new position
            if shrink == "max":
                p1 = sampler.flatchain[np.argmax(sampler.flatlnprobability)]
            else:
                p1 = np.median(pos, axis=0)
            pos_std = np.std(sampler.flatchain, axis=0) * 0.5

            # reset sampler
            sampler.reset()

            # randomly generate new start position
            pos0 = rand_pos(p1, nwalkers=nwalkers, eps=pos_std)

    else:
        # [o] single burn-in
        # run mcmc
        print("@Regli: running burn-in [{}]...".format(n_burnin))
        pos, prob, state = sampler.run_mcmc(pos0, n_burnin)

        # shrink to a new position
        if shrink == "max":
            p1 = sampler.flatchain[np.argmax(sampler.flatlnprobability)]
        else:
            p1 = np.median(pos, axis=0)
        pos_std = np.std(sampler.flatchain, axis=0) * 0.5

        # reset sampler
        sampler.reset()

        # randomly generate new start position
        pos0 = rand_pos(p1, nwalkers=nwalkers, eps=pos_std)

    # run mcmc
    print("@Regli: running chains [{}]...".format(n_step))
    pos, prob, state = sampler.run_mcmc(pos0, n_step)

    if full:
        # return full result
        return sampler, pos, prob, state
    else:
        # return sampler only
        return sampler